import math
import tensorflow as tf
import pathlib
import numpy as np
from matplotlib import pyplot
from keras.models import Model
import sys
import io

sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')

def output_model(the_model):
  w = 24
  l = len(the_model.layers)
  ll = 'Layer:'.ljust(w)
  il = 'Input Shape:'.ljust(w)
  ol = 'Output Shape'
  print(ll + il + ol)
  for i in range(l):
    n = type(the_model.layers[i]).__name__.ljust(w)
    ish = str(the_model.layers[i].input.shape).ljust(w)
    osh = str(the_model.layers[i].output.shape)
    print(n +  ':' + ish +  ':' + osh)

def getEventNumber(ev):
# REPLACE NEXT
  concatOffsets = [CONCAT_OFFSETS]
  if len(concatOffsets) == 0:
    return ev
  evt = ev
  for offset in concatOffsets:
    if evt - offset > 0:
      evt = evt - offset
    else:
      return evt
  return evt

def getImageFilePrefix(ev):
# REPLACE NEXT
  prefixes = [IMAGE_PREFIXES]
# REPLACE NEXT
  concatOffsets = [CONCAT_OFFSETS]
  pLen = len(prefixes)
  if pLen == 0:
    return 'UNKNOWN'
  evt = ev
  i = 0
  current = prefixes[i]
  for offset in concatOffsets:
    if evt - offset > 0:
      evt = evt - offset
      i = i + 1
      if (i < pLen):
        current = prefixes[i]
    else:
      return current
  return current

def showConvolutionFeatures(img_array, model, outputFolder):
  numCols = 8
  debug = False
  layer_names = [ layer.name for layer in model.layers ]
  layer_outputs = [ layer.output for layer in model.layers ]
  feature_map_model = Model(inputs=model.inputs, outputs=layer_outputs)
  feature_maps = feature_map_model.predict(img_array)
  for i in range(len(feature_maps)):
    fmap = feature_maps[i]
    layerName = layer_names[i]
    layerType = type(model.layers[i]).__name__
    if (debug):
      print(layerName, end = '')
      print(' layerType=', end = '')
      print(layerType)
      print(' fmap.shape=', end = '')
      print(fmap.shape)
    if 'Conv2D' not in layerType:
      continue
    if len(fmap.shape) != 4:
      continue
    ix = 1
    lastDim = fmap.shape[-1]
    numRows = int(lastDim / numCols)
    if (numRows < 1):
      continue
    fname = outputFolder + '.' + layerType + '.' + str(i) + '.plot.png'
    for ii in range(lastDim):
      ax = pyplot.subplot(numRows, numCols, ix)
      ax.set_xticks([])
      ax.set_yticks([])
      feature_image = fmap[0, :, :, ii]
      pyplot.imshow(feature_image, cmap='gray')
      ix += 1
    pyplot.savefig(fname)
    if (debug):
      pyplot.show()


# REPLACE NEXT
img_height = IMAGE_HEIGHT
# REPLACE NEXT
img_width = IMAGE_WIDTH

numPerFolder = 10000

try:

# REPLACE NEXT
  model = tf.keras.models.load_model('SAVED_MODEL')
# model.summary()
  output_model(model)

  testImages = [ ]
  eventNums = []
# REPLACE NEXT
  prefix = "IMAGE_DIR"
# REPLACE NEXT
  imageExt = ".IMAGE_EXTENSION"

# REPLACE NEXT
  class_names = [ CLASS_NAMES ]

# REPLACE NEXT
  ev = EXPLAIN_EVENT_NUM
  ev = getEventNumber(ev)
  evStr = str(ev)
  ev = math.floor(ev / numPerFolder) * numPerFolder
# REPLACE NEXT
  sub = "%0EVENT_FIELD_LENGTHd" % ev
# REPLACE NEXT
  sub = sub.zfill(EVENT_FIELD_LENGTH)
# REPLACE NEXT
  evStr = evStr.zfill(EVENT_FIELD_LENGTH)

# REPLACE NEXT
  imagePath = prefix + getImageFilePrefix(ev) + evStr + "IMAGE_SUFFIX" + imageExt
  imageFile = pathlib.Path(imagePath)

  img = tf.keras.utils.load_img( imageFile, target_size=(img_height, img_width) )
  img_array = tf.keras.utils.img_to_array(img)
  img_array = tf.expand_dims(img_array, 0) # Create a batch
  outFolder = prefix + evStr
  showConvolutionFeatures(img_array, model, outFolder)

  predictions = model.predict(img_array)
  score = predictions[0]

  predFileName = outFolder + '.prediction.xml'
  outFile = open(predFileName, "w")
  outFile.write("<PredictionInfo>")
  for j in range(len(score)):
    outFile.write(str(score[j] * 100) + ", ")
  outFile.write("\n")
  outFile.write("</PredictionInfo>")
  outFile.close()

except Exception as inst:
  print('Exception in explanation script')
  print(type(inst))
  print(inst.args)
  print(inst)

else:
  print('Explanation script completed')
