import matplotlib.pyplot as plt
import numpy as np
import os
import PIL
import tensorflow as tf
import sys
import io

sys.stdout = io.TextIOWrapper(sys.stdout.buffer, encoding='utf-8')

from tensorflow import keras
from tensorflow.keras import layers
from tensorflow.keras.models import Sequential

import pathlib

def output_model(the_model):
  w = 24
  l = len(the_model.layers)
  ll = 'Layer:'.ljust(w)
  il = 'Input Shape:'.ljust(w)
  ol = 'Output Shape'
  print(ll + il + ol)
  for i in range(l):
    n = type(the_model.layers[i]).__name__.ljust(w)
    ish = str(the_model.layers[i].input.shape).ljust(w)
    osh = str(the_model.layers[i].output.shape)
    print(n +  ':' + ish +  ':' + osh)

def plot_curves(the_history, key1, key2, leg1, leg2, xlabel, ylabel, title):
  plt.figure(figsize=[6,4])
  plt.plot(the_history.history[key1],'r',linewidth=2.0)
  plt.plot(the_history.history[key2],'b',linewidth=2.0)
  plt.legend([leg1, leg2],fontsize=10)
  plt.xlabel(xlabel,fontsize=10)
  plt.ylabel(ylabel,fontsize=10)
  plt.title(title,fontsize=10)

def writeModelXML(model, outputFile, historyMap):
  acc = historyMap['accuracy']
  val_acc = historyMap['val_accuracy']
  loss = historyMap['loss']
  val_loss = historyMap['val_loss']

  outFile = open(outputFile, "w")
  outFile.write("<ModelRunInfo accuracy=\"{}\" valAccuracy=\"{}\" ".format(acc[epochs-1], val_acc[epochs-1]))
  outFile.write(" loss=\"{}\" valLoss=\"{}\" >".format(loss[epochs-1], val_loss[epochs-1]))

  for i in range(len(model.layers)):
    layer = model.layers[i]
    layerName = layer.name
    layerType = type(layer).__name__
    outFile.write("<ModelLayer name=\"")
    outFile.write(layerName)
    outFile.write("\" type=\"")
    outFile.write(layerType)
    outFile.write("\" />")

  outFile.write("</ModelRunInfo>")
  outFile.close()


# Change path to one that works on your computer
# REPLACE NEXT
data_dir = pathlib.Path('TRAINING_DIR')
# print(data_dir)

# REPLACE NEXT
image_count = len(list(data_dir.glob('**/*/*.IMAGE_EXTENSION')))
# print(image_count)

batch_size = 32
# REPLACE NEXT
img_height = IMAGE_HEIGHT
# REPLACE NEXT
img_width = IMAGE_WIDTH

train_ds = tf.keras.utils.image_dataset_from_directory(
  data_dir,
  validation_split=0.2,
  subset="training",
  seed=123,
  image_size=(img_height, img_width),
  batch_size=batch_size)

val_ds = tf.keras.utils.image_dataset_from_directory(
  data_dir,
  validation_split=0.2,
  subset="validation",
  seed=123,
  image_size=(img_height, img_width),
  batch_size=batch_size)

class_names = train_ds.class_names
print('CLASS NAMES')
print(class_names)

data_augmentation = Sequential([
  layers.RandomFlip(seed=123),
  layers.RandomRotation(0.1, seed=123),
  layers.RandomTranslation(height_factor=(-0.2, 0.3), width_factor=(-0.2, 0.3), seed=123)
])

AUTOTUNE = tf.data.AUTOTUNE

train_ds = train_ds.map(
  lambda x, y: (data_augmentation(x, training=True), y))

train_ds = train_ds.cache().shuffle(1000).prefetch(buffer_size=AUTOTUNE)
val_ds = val_ds.cache().prefetch(buffer_size=AUTOTUNE)

num_classes = len(class_names)

# REPLACE NEXT
MODEL_CREATION

print('# REPORT')
output_model(model)
# model.summary()
print('# REPORT')

# REPLACE NEXT
epochs=NUM_EPOCHS
history = model.fit(
  train_ds,
  validation_data=val_ds,
  epochs=epochs
)

acc = history.history['accuracy']
val_acc = history.history['val_accuracy']

loss = history.history['loss']
val_loss = history.history['val_loss']

# REPLACE NEXT
model.save('SAVED_MODEL')

# REPLACE NEXT
writeModelXML(model, "MODEL_RUN_INFO", history.history)

#Plot the Loss Curves
plot_curves(history, 'loss', 'val_loss', 'Training Loss', 'Validation Loss', 'Epochs', 'Loss', 'Loss Curves')
# REPLACE NEXT
plt.savefig('RUN_DIRlossCurves.MODEL_NAME.png')

#Plot the Accuracy Curves
plot_curves(history, 'accuracy', 'val_accuracy', 'Training Accuracy', 'Validation Accuracy', 'Epochs', 'Accuracy', 'Accuracy Curves')
# REPLACE NEXT
plt.savefig('RUN_DIRaccuracyCurves.MODEL_NAME.png')
